/*
    fsext.c
    Copyright 1999 by Richard Dawe

    Perform some behind-the-scenes trickery with DJGPP's File System Extensions
    so that file handles can be used for (one-way) file operations.

    ---

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <io.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <dir.h>

#include <sys/fsext.h>

#include <mailslot.h>
#include "internal.h"

typedef struct _MAILSLOT_FSEXT {
    int   fd;
    int   handle;
    char *name;
} MAILSLOT_FSEXT;

/* We need this to keep track of which file descriptors are mailslots. */
MAILSLOT_FSEXT *__mailslot_fsext[MAILSLOT_MAX];

int __mailslot_fsext_inited = 0;

int  __mailslot_install_handler (void);
void __mailslot_uninstall_handler (void);
int  __mailslot_handler (__FSEXT_Fnumber n, int *rv, va_list args);

/* ------------------------------
   - __mailslot_install_handler -
   ------------------------------ */

int __mailslot_install_handler (void)
{
    int i;

    if (__mailslot_fsext_inited) return(__mailslot_fsext_inited);

    for (i = 0; i < MAILSLOT_MAX; i++) { __mailslot_fsext[i] = NULL; }
    __FSEXT_add_open_handler(__mailslot_handler);
    atexit(__mailslot_uninstall_handler);
    __mailslot_fsext_inited = 1;

    return(__mailslot_fsext_inited);
}

/* ------------------------------
   - __mailslot_install_handler -
   ------------------------------ */

void __mailslot_uninstall_handler (void)
{
    /* TODO: Free memory & uninstall open handler */
}

/* ------------------------------
   - __mailslot_fsext_firstfree -
   ------------------------------ */

int __mailslot_fsext_firstfree (void)
{
    int i, msd;

    for (i = 0, msd = -1; i < MAILSLOT_MAX; i++) {
        if (__mailslot_fsext[i] == NULL) {
            msd = i;
            break;
        }
    }

    return(msd);
}            

/* ----------------------
   - __mailslot_handler -
   ---------------------- */

int __mailslot_handler (__FSEXT_Fnumber n, int *rv, va_list args)
{
    int emul = 0;       /* Emulated the call? */
    char *file = NULL;
    int mode = 0, perm = 0;
    int msd, hMailslot, fd;
    MAILSLOT_INFO mi;
    char *p = NULL, *q = NULL, *buf = NULL;
    char newname[MAXPATH];
    int len;
    int i;

    switch(n)
    {
        /* --- __FSEXT_creat, __FSEXT_open --- */
        case __FSEXT_creat:
        case __FSEXT_open:
            /* Get the parameters */
            file = strdup((char *) va_arg(args, const char *));

            if (n ==  __FSEXT_creat) {
                perm = va_arg(args, int);

                /* Fake the mode flag based on the permissions. */
                mode = ((perm & S_IWUSR) == S_IWUSR) ? O_WRONLY : O_RDONLY;
            } else {
                mode = va_arg(args, int);
                perm = va_arg(args, int);
            }

            /* Forward-slashify file */
            for (i = 0; i < strlen(file); i++) {
                if (file[i] == '\\') file[i] = '/';
            }

            /* Refers to a mailslot? */
            if (strstr(file, MAILSLOT_DEV) != file) {
                free(file);
                break;
            }

            emul = 1;   /* Emulated */

            /* What kind of a mailslot in needed? */
            /* TODO: Handle other modes too, i.e. text/binary. */

            /* NB: We have to use file descriptors in binary mode. write()
               doesn't call the FSEXT unless O_BINARY is set. Then write()
               calls _write(), which in turn calls the FSEXT. This problem
               is fixed in DJGPP 2.02. Unfortunately, setmode() here has no
               effect since the open() function will set the mode after
               the FSEXT call has completed. */

            if ((mode & O_RDWR) == O_RDWR) {
                /* Only read-only or write-only allowed - uni-directional */
                *rv = -1;
                errno = EPERM;  /* TODO: Error code? */
            } else if ((mode & O_WRONLY) == O_WRONLY) {
                *rv = -1;   /* Fail by default */

                /* Store the mailslot details for write calls */
                msd = __mailslot_fsext_firstfree();
                p = file + strlen(MAILSLOT_DEV);

                if (msd == -1) {
                    errno = EAGAIN; /* Should be free sometime soon */
                } else {
                    /* Deduce the actual mailslot path */
                    bzero(newname, sizeof(newname));

                    if (strstr(p, MAILSLOT_LOCAL) == p) {
                        /* Local mailslot */
                        p += strlen(MAILSLOT_LOCAL);
                        sprintf(newname, "/mailslot/%s", p);
                    } else if (strstr(p, MAILSLOT_BROADCAST) == p) {
                        /* Broadcast mailslot */
                        p += strlen(MAILSLOT_BROADCAST);
                        sprintf(newname, "//*/mailslot/%s", p);
                    } else {
                        /* Remote mailslot */
                        sprintf(newname, "//%s", p);
                        q = strchr(newname + 2, '/');
                        if (q != NULL) *q = '\0';

                        p = strchr(p, '/');
                        if (p != NULL) {
                            strcat(newname, "/mailslot");
                            strcat(newname, p);
                        }
                    }
                    
                    __mailslot_fsext[msd] = (MAILSLOT_FSEXT *) malloc(sizeof(MAILSLOT_FSEXT));
                    if (__mailslot_fsext[msd] != NULL) {
                        fd = __FSEXT_alloc_fd(__mailslot_handler);
                        __mailslot_fsext[msd]->fd     = fd;
                        __mailslot_fsext[msd]->handle = -1;
                        __mailslot_fsext[msd]->name   = strdup(newname);
                        *rv = fd;
                    }
                }
            } else if ((mode & O_RDONLY) == O_RDONLY) {
                *rv = -1;   /* Fail by default */

                /* Create a mailslot */
                p = file + strlen(MAILSLOT_DEV);

                /* Deduce the actual mailslot path */
                bzero(newname, sizeof(newname));

                if (strstr(p, MAILSLOT_LOCAL) == p) {
                    /* Local mailslot */
                    p += strlen(MAILSLOT_LOCAL);
                    sprintf(newname, "/mailslot/%s", p);
                } else {
                    /* Baaad filename */
                    errno = EINVAL;
                }

                hMailslot = __mailslot_create(newname, MAILSLOT_MAXSIZE);

                if (hMailslot != -1) {
                    msd = __mailslot_fsext_firstfree();

                    if (msd == -1) {
                        errno = EAGAIN; /* Should be free sometime soon */
                    } else {
                        __mailslot_fsext[msd] = (MAILSLOT_FSEXT *) malloc(sizeof(MAILSLOT_FSEXT));
                        if (__mailslot_fsext[msd] != NULL) {
                            fd = __FSEXT_alloc_fd(__mailslot_handler);
                            __mailslot_fsext[msd]->fd     = fd;
                            __mailslot_fsext[msd]->handle = hMailslot;
                            __mailslot_fsext[msd]->name   = NULL;
                            *rv = fd;
                        }
                    }
                }
            }

            free(file);
            break;

        /* --- __FSEXT_close --- */
        case __FSEXT_close:
            *rv = -1;   /* Fail by default */

            /* Find the data structure */
            fd = va_arg(args, int);
            for (i = 0, msd = -1; i < MAILSLOT_MAX; i++) {
                if (__mailslot_fsext[i]->fd == fd) {
                    msd = i;
                    break;
                }
            }

            if (msd == -1) {
                break;
            } else {
                __FSEXT_set_function(fd, NULL);
                _close(fd);

                if (__mailslot_fsext[msd]->handle != -1)
                    __mailslot_destroy(__mailslot_fsext[msd]->handle);
                if (__mailslot_fsext[msd]->name   != NULL)
                    free(__mailslot_fsext[msd]->name);

                free(__mailslot_fsext[msd]);
                __mailslot_fsext[msd] = NULL;
                *rv = 0;
            }

            emul = 1;
            break;

        /* --- __FSEXT_write --- */
        case __FSEXT_write:
            *rv = -1;   /* Fail by default */

            fd  =          va_arg(args, int);
            buf = (char *) va_arg(args, void *);
            len = (int)    va_arg(args, size_t);

            /* Find the data structure */
            for (i = 0, msd = -1; i < MAILSLOT_MAX; i++) {
                if (__mailslot_fsext[i]->fd == fd) {
                    msd = i;
                    break;
                }
            }

            if (msd == -1) break;

            /* OK, this is a mailslot, so emulate the call */
            emul = 1;

            if (__mailslot_fsext[i]->name == NULL) {
                errno = EPERM;
                break;
            }

            /* TODO: Split the buffer up for writing since we can only write
               in small chunks. */
            *rv = __mailslot_write(__mailslot_fsext[i]->name, buf, len, 0);
            break;

        /* --- __FSEXT_read --- */
        case __FSEXT_read:
            *rv = -1;   /* Fail by default */

            fd  =          va_arg(args, int);
            buf = (char *) va_arg(args, void *);
            len = (int)    va_arg(args, size_t);

            /* Find the data structure */
            for (i = 0, msd = -1; i < MAILSLOT_MAX; i++) {
                if (__mailslot_fsext[i]->fd == fd) {
                    msd = i;
                    break;
                }
            }

            if (msd == -1) break;

            /* OK, this is a mailslot, so emulate the call */
            emul = 1;

            if (__mailslot_fsext[i]->handle == -1) {
                errno = EPERM;
                break;
            }

            /* TODO: Read the buffer in small chunks. */
            *rv = __mailslot_read(__mailslot_fsext[i]->handle, buf, len, 0);
            break;

        /* --- __FSEXT_ready --- */
        case __FSEXT_ready:
            *rv = 0;    /* Not ready by default */

            /* Find the data structure */
            for (i = 0, msd = -1; i < MAILSLOT_MAX; i++) {
                if (__mailslot_fsext[i]->fd == fd) {
                    msd = i;
                    break;
                }
            }

            if (msd == -1) break;
            emul = 1;

            /* Write-only mailslots always ready for writing. */
            if (__mailslot_fsext[i]->name != NULL) *rv |= __FSEXT_ready_write;

            /* Read-only mailslots need checking. */
            if (__mailslot_fsext[i]->handle != -1) {
                if (__mailslot_info(__mailslot_fsext[i]->handle, &mi) == 0) {
                    if (mi.queued > 0) *rv |= __FSEXT_ready_read;
                }
            }

            /* TODO: select() on error conditions? */
            break;

        default:
            break;
    }

    return(emul);
}
