/*
    syslog.c
    Copyright 1999 by Richard Dawe

    openlog(3), syslog(3), vsyslog(3) and closelog(3) functions.

    ---

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>
#include <sys/param.h>

#include <syslog.h>
#include <mailslot.h>

#define SYSLOG_DEFAULT_FACILITY LOG_USER

int   __syslog_opened   = 0;
char *__syslog_ident    = NULL;
int   __syslog_option   = 0;
int   __syslog_facility = SYSLOG_DEFAULT_FACILITY;
FILE *__syslog_fp       = NULL;

/* -----------
   - openlog -
   ----------- */

/* 'option' needs to be handled in the following way:

   . LOG_NDELAY is ignored, because a connection is never opened for writing
     directly.

   . LOG_PID is emulated, by including the PID with each message in syslog(),
     rather than syslogd adding it. This is because syslogd will not be able
     to get the PID.

   . LOG_CONS, LOG_PERROR are handled as usual.
*/

void openlog (const char *ident, int option, int facility)
{
    /* Install the mailslot handler */
    __mailslot_install_handler();

    /* Re-opening without a closelog()? */
    if (__syslog_opened) {
        fclose(__syslog_fp);
        free(__syslog_ident);
        __syslog_ident = NULL;
    }

    __syslog_ident    = strdup(ident);    
    __syslog_option   = option;
    __syslog_facility = facility;
    __syslog_fp       = fopen("/dev/mailslot/localhost/syslogd", "wb");
    __syslog_opened   = 1;
}

/* ------------
   - closelog -
   ------------ */

void closelog (void)
{
    if (__syslog_opened) {
        fclose(__syslog_fp);
        free(__syslog_ident);
        __syslog_ident = NULL;
    }

    __syslog_opened   = 0;
    __syslog_option   = 0;
    __syslog_facility = SYSLOG_DEFAULT_FACILITY;
}

/* ----------
   - syslog -
   ---------- */

void syslog (int pri, const char *fmt, ...)
{
    va_list args = (va_list) ((&fmt) + 1);
    vsyslog(pri, fmt, args);
}

/* -----------
   - vsyslog -
   ----------- */

void vsyslog (int pri, const char *fmt, va_list args)
{
    char    *newfmt = NULL;
    int      len    = 0;
    time_t   now;
    char    *cnow   = NULL;
    char    *p      = NULL;
    char    *q      = NULL;
    char     hnbuf[MAXGETHOSTNAME];
    char     itoabuf[33];

    /* Open the log first, if necessary */
    if (!__syslog_opened) openlog(NULL, 0, SYSLOG_DEFAULT_FACILITY);

    /* --- Construct the prefix --- */

    /* Get the time */
    time(&now);
    cnow = ctime((const time_t *) &now);
    p = strstr(cnow, "\n");
    if (p != NULL) *p = '\0';   /* Nuke the newline in the time */

    len = strlen(cnow) + 1 + 1; /* null + separator */

    /* Get the hostname */
    if (gethostname(hnbuf, sizeof(hnbuf)) != 0) strcpy(hnbuf, "localhost");
    p = strchr(hnbuf, ' ');
    if (p != NULL) *p = '\0';

    len += strlen(hnbuf) + 1 + 1; /* null + separator */

    /* Add the priority number */
    bzero(itoabuf, sizeof(itoabuf));
    itoa(pri, itoabuf, 10);
    len += strlen(itoabuf) + 2; /* angled brackets */

    /* Add the PID if option LOG_PID given */
    /* 10 digits (max) + 2 square brackets */
    if ((__syslog_option & LOG_PID) == LOG_PID) len += 12;

    len += 2;           /* colon + space */
    len += strlen(fmt);

    /* Add the current error message(s) */
    for (p = (char *) fmt; (p = strstr(p, "%m")) != NULL; ) {
        p   += 2;
        len += strlen(strerror(errno));
    }

    /* Terminating newline needed? */
    if (strstr(fmt, "\n") != (fmt + strlen(fmt) - strlen("\n") + 1))
        len += 1;

    /* Build newfmt */
    newfmt = (char *) malloc(len);
    if (newfmt != NULL) {
        sprintf(newfmt, "%s %s <%s>", cnow, hnbuf, itoabuf);

        if ((__syslog_option & LOG_PID) == LOG_PID)
            sprintf(newfmt, "%s[%i]", newfmt, getpid());

        sprintf(newfmt, "%s: %s", newfmt, fmt);

        for (p = newfmt; (p = strstr(p, "%m")) != NULL; ) {
            *p = '\0';
            p  += 2;
            q  =  strdup(p);                
            strcat(newfmt, strerror(errno));
            strcat(newfmt, q);
            free(q);
        }

        if (strstr(fmt, "\n") != (fmt + strlen(fmt) - strlen("\n") + 1))
            strcat(newfmt, "\n");
    } else {
        /* Fallback */
        newfmt = (char *) fmt;
    }

    /* Log it */
    if (__syslog_fp != NULL) _doprnt(newfmt, args, __syslog_fp);

    /* Display to console on error, if LOG_CONS option used. */
    if (ferror(__syslog_fp) && ((__syslog_option & LOG_PERROR) == LOG_CONS))
        _doprnt(newfmt, args, stdout);

    /* Display on stderr too, if LOG_PERROR option used. */
    if ((__syslog_option & LOG_PERROR) == LOG_PERROR)
        _doprnt(newfmt, args, stderr);

    if (newfmt != fmt) free(newfmt);
}
